'***************************************************
'* 0.1 - Spin code generated by cspin version 0.43 *
'* 0.2 - small changes to remove unneeded vars     *
'* 0.3 - comments and cleanup                      *
'* 0.4 - added initQueHead, getFirst, getNext      *
'***************************************************

{ 
  insque(3C) routines
    This file started as public domain C code.


NAME
        insque, remque -- insert, remove an element from a queue
SYNOPSIS
        struct qelem {
          struct qelem *q_forw;
          struct qelem *q_back;
          char q_data[];
        };
        void insque (struct qelem *elem, struct qelem *pred)
        void remque (struct qelem *elem)

DESCRIPTION
        Routines to manipulate queues built from doubly linked lists.

        The insque routine inserts ELEM in the queue immediately after PRED.  Use
        PRED as the queue head.

        The remque routine removes ELEM from its containing queue.

        Elements using these routines must reserve the first eight bytes for the 
        forward and backward pointers.
}

  ' struct qelem
  q_forw = 0    ' not needed but use this to make code more clear
  q_back = 2
  q_data = 4
  
' 
PUB insque(elem, pred, semID)
  '/**
  ' * insert elem
  ' * insert elem after pred
  ' * if semID is not -1 then use it as a lock
  ' *
  ' * @return None
  ' */
  
  if (semID <> -1)
    repeat until not lockset(semID)

  ' 1. set the forward ptr of new element to point to what the pred node was pointing at
  WORD[elem + q_forw] := WORD[pred + q_forw]

  ' 2. set the back pointer that pred was pointing at point to the new element
  WORD[WORD[pred + q_forw] + q_back] := elem

  ' 3. set the new element point back to the pred
  WORD[elem + q_back] := pred

  ' 4. set pred point forward to the new element
  WORD[pred + q_forw] := elem

  if (semID <> -1)
    lockclr(semID)


PUB insqueTail(elem, head, semID) | pred
  '/**
  ' * insert elem at the end of the queue
  ' * if semID is not -1 then use it as a lock
  ' *
  ' * @return None
  ' */

  if (semID <> -1)
    repeat until not lockset(semID)
  
  pred := WORD[head + q_back]
  
  ' 1. set the forward ptr of new element to point to what the pred node was pointing at
  WORD[elem + q_forw] := WORD[pred + q_forw]

  ' 2. set the back pointer that pred was pointing at point to the new element
  WORD[WORD[pred + q_forw] + q_back] := elem

  ' 3. set the new element point back to the pred
  WORD[elem + q_back] := pred

  ' 4. set pred point forward to the new element
  WORD[pred + q_forw] := elem

  if (semID <> -1)
    lockclr(semID)

PUB remque(elem, semID)
  '/**
  ' * remove elem
  ' * remove elem from queue
  ' * if semID is not -1 then use it as a lock
  ' *
  ' * @return None
  ' */
  if (semID <> -1)
    repeat until not lockset(semID)

  ' 1. set the back pointer of the node that elem was forward pointing to to what elem is back pointing to
  WORD[WORD[elem + q_forw] + q_back] := WORD[elem + q_back]

  ' 2. 
  WORD[WORD[elem + q_back] + q_forw] := WORD[elem + q_forw]
  
  if (semID <> -1)
    lockclr(semID)


PUB initQueHead(queHead)
  '/**
  ' * initialize the queue header
  ' *
  ' * @return None
  ' */
  WORD[queHead + q_forw] := queHead
  WORD[queHead + q_back] := queHead

PUB getFirst(queHeadPtr, semID) : retPtr
  '/**
  ' * get first element
  ' * return the first element of the list
  ' * if semID is not -1 then use it as a lock
  ' *
  ' * @return address of first element or 0 if list is empty
  ' */
  if (semID <> -1)
    repeat until not lockset(semID)

  retPtr := WORD[queHeadPtr + q_forw]
  if (retPtr == queHeadPtr)
    retPtr := 0
 
  if (semID <> -1)
    lockclr(semID)

PUB getNext(queHeadPtr, ptr, semID) : retPtr
  '/**
  ' * get next element
  ' * return the element after ptr
  ' * if semID is not -1 then use it as a lock
  ' *
  ' * @return address of next element or 0 if list is empty
  ' */
  if (semID <> -1)
    repeat until not lockset(semID)

  retPtr := WORD[ptr + q_forw] 
  if (retPtr == queHeadPtr)
    retPtr := 0
 
  if (semID <> -1)
    lockclr(semID)
    
    